%{
#include "minako.h"
#include <stdlib.h>
#include <assert.h>
	
#define YYERROR_VERBOSE of course!

%}

/* definiere einen Zusatzparameter (die Symboltabelle) */
%parse-param{void* table}

%union {
	double floatValue;
	int intValue;
	char* string;
}

/* benutzte Terminalsymbole */
%token AND
%token OR
%token EQ
%token NEQ
%token LEQ
%token GEQ
%token LSS
%token GRT
%token KW_BOOLEAN
%token KW_DO
%token KW_ELSE
%token KW_FLOAT
%token KW_FOR
%token KW_IF
%token KW_INT
%token KW_PRINTF
%token KW_RETURN
%token KW_VOID
%token KW_WHILE
%token <intValue> CONST_INT
%token <floatValue> CONST_FLOAT
%token <intValue> CONST_BOOLEAN
%token <string> CONST_STRING
%token <string> ID

/* Definition der Operatorassoziativitäten und -präzedenzen */
%nonassoc LOWER_THAN_ELSE
%nonassoc KW_ELSE
%left AND
%left OR
%left EQ NEQ GEQ LEQ GRT LSS
%left '+' '-'
%left '*' '/'
%nonassoc UMINUS

%%

program:
	/* leer */
	| program declassignment ';'
	| program functiondefinition
	;

functiondefinition:
	type ID '(' opt_formalparameter ')' '{' statementlist '}'
	;

opt_formalparameter:
	/* leer */
	| type ID lop_formalparameter
	;

lop_formalparameter:
	/* leer */
	| ',' type ID lop_formalparameter
	;

functioncall:
	ID '(' ')'
	| ID '(' assignment lop_actualparameter ')'
	;
	
lop_actualparameter:
	/* leer */
	| ',' assignment lop_actualparameter
	;

statementlist:
	/* leer */
	| statementlist block
	;

block: 
	'{' statementlist '}' 
	| statement
	;

block2: 
	'{' statementlist '}' 
	| statement
	;

statement: 
	  ifstatement 
	| forstatement 
	| whilestatement 
	| returnstatement ';' 
	| dowhilestatement ';' 
	| printf ';' 
	| declassignment ';' 
	| statassignment ';' 
	| functioncall ';'
	;

ifstatement: 
	KW_IF '(' assignment ')' block2 %prec LOWER_THAN_ELSE
	| KW_IF '(' assignment ')' block2 KW_ELSE block2
	;

forstatement: 
	KW_FOR '(' declassignment ';' expr ';' statassignment ')' block2 
	| KW_FOR '(' statassignment ';' expr ';' statassignment ')' block2 
	;

dowhilestatement: 
	KW_DO block2 KW_WHILE '(' assignment ')'
	;

whilestatement: 
	KW_WHILE '(' assignment ')' block2 
	;

returnstatement: 
	KW_RETURN
	| KW_RETURN assignment
	;

printf: 
	KW_PRINTF '(' assignment ')' 
	| KW_PRINTF '(' CONST_STRING ')'
	;

declassignment: 
	type ID
	| type ID '=' assignment
	;

type:
	KW_BOOLEAN
	| KW_FLOAT
	| KW_INT
	| KW_VOID
	;

statassignment: 
	ID '=' assignment
	;

assignment:
	ID '=' assignment
	| expr
	;

expr:
	expr '+' expr
	| expr '-' expr
	| expr '*' expr
	| expr '/' expr
	| '-' expr %prec UMINUS
	| expr AND expr
	| expr OR expr
	| expr EQ expr
	| expr NEQ expr
	| expr LEQ expr
	| expr GEQ expr
	| expr LSS expr
	| expr GRT expr
	| CONST_INT
	| CONST_FLOAT
	| CONST_BOOLEAN
	| functioncall
	| ID
	| '(' assignment ')'
	;

%%

int main(int c, char** v)
{
	symtab* table = newSymtab(0);
	yydebug = c-1;
	return yyparse((void*) table);
}

/* Die folgenden Funktionen dienen der vereinfachten Benutzung der
 * Symboltabelle. Sie dürfen beliebig verändert oder erweitert werden. */

#define NEW(_struct) (_struct*)malloc(sizeof(_struct))
	
/* ********************************************************* Minako Variablen */
	
	symtabSymbol* newMinakoVariable(long type, const char* id)
	{
		symtabSymbol* sym = newSymtabSymbol('var', id);
		minakoVariable* var = NEW(minakoVariable);
		
		var->type = type;
		var->line = yylineno;
		
		sym->data = var;
		
		return sym;
	}
	
	void minakoVariableDeclare(symtab* table, const symtabSymbol* symb)
	{
		assert(symb && symb->class == 'var');
		
		/* teste, ob die Variable schon existiert */
		if (!symtabDeclare(table, symb))
		{
			/* suche das Originalsymbol für eine schönere Fehlermeldung */
			symb = symtabGet(table, symb->id);
			
			switch (symb->class)
			{
			case 'var':
				yyerror("Die Variable \"%s\" wurde bereits in Zeile %i deklariert.",
				        symb->id, ((minakoVariable*)symb->data)->line);
				
			case 'func':
				yyerror("Es existiert bereits eine Funktion in Zeile %i mit dem Namen "
				        "\"%s\".", ((minakoFunction*)symb->data)->line, symb->id);
				
			default:
				yyerror("Der Bezeichner \"%s\" ist bereits vergeben.", symb->id);
			}
		}
	}
	
/* ******************************************************** Minako Funktionen */
	
	symtabSymbol* newMinakoFunction(long type, const char* id)
	{
		symtabSymbol* symb = newSymtabSymbol('func', id);
		minakoFunction* func = NEW(minakoFunction);
		
		func->type = type;
		func->line = yylineno;
		
		symb->data = func;
		symb->destructor = (fDestructor) deleteMinakoFunction;
		
		return symb;
	}
	
	void deleteMinakoFunction(minakoFunction* func)
	{
		int i = func->argCount;
		
		while (i--)
			deleteSymtabSymbol(func->param[i]);
		
		free(func->param);
		free(func);
	}
	
	void minakoFunctionDeclare(symtab* table, const symtabSymbol* symb,
	                           stack* parameter)
	{
		minakoFunction* func;
		stack* p;
		int count;
		
		assert(symb && symb->class == 'func' && parameter);
		
		func = (minakoFunction*) symb->data;
		
		/* prüfe, ob die Funktion schon in der Symboltabelle existiert */
		if (!symtabDeclare(table, symb))
		{
			/* für eine schönere Fehlerausgabe, suche das Originalsymbol */
			symb = symtabGet(table, symb->id);
			
			switch (symb->class)
			{
			case 'var':
				yyerror("Es existiert bereits eine Variable in Zeile %i mit dem Namen "
				        "\"%s\".", ((minakoVariable*)symb->data)->line, symb->id);
				
			case 'func':
				yyerror("Die Funktion \"%s\" wurde bereits in Zeile %i deklariert.",
				        symb->id,	((minakoFunction*)symb->data)->line);
				
			default:
				yyerror("Der Bezeichner \"%s\" ist bereits vergeben.", symb->id);
			}
		}
		
		/* zähle die Elemente auf dem Stack */
		p = newStack();
		count = 0;
		
		while (!stackIsEmpty(parameter))
		{
			stackPush(p, stackPop(parameter));
			++count;
		}
		
		deleteStack(parameter);
		
		/* nachdem die Anzahl bekannt ist, ist ein statisches Array günstiger */
		func->param = (symtabSymbol**) malloc(sizeof(symtabSymbol*) * count);
		func->argCount = 0;
		
		while (count--)
		{
			symtabSymbol* s = (symtabSymbol*)stackPop(p);
			func->param[func->argCount++] = s;
			
			/* Destruktor auf 0 setzen führt dazu, dass die Parameter nicht von der
			 * Symboltabelle freigegeben werden - wir kümmern uns selbst darum */
			s->destructor = 0;
		}
		
		deleteStack(p);
	}
	
#undef NEW
